# Copyright (c) 2010, 2011, Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Read / Write an OOPS dict as an rfc822 formatted message.

This style of OOPS format is very web server specific, not extensible - it
should be considered deprecated.

The reports this serializer handles always have the following variables:

* id: The name of this error report.
* type: The type of the exception that occurred.
* value: The value of the exception that occurred.
* time: The time at which the exception occurred.
* pageid: The identifier for the template/script that oopsed.
* branch_nick: The branch nickname.
* revno: The revision number of the branch.
* tb_text: A text version of the traceback.
* username: The user associated with the request.
* url: The URL for the failed request.
* req_vars: The request variables.
* branch_nick: A name for the branch of code that was running when the report
  was triggered.
* revno: The revision that the branch was at.
* Informational: A flag, True if the error wasn't fatal- if it was
  'informational'.
"""


__all__ = [
    'read',
    'write',
    ]

__metaclass__ = type

import datetime
import logging
import rfc822
import re
import urllib

import iso8601


def read(fp):
    """Deserialize an OOPS from an RFC822 format message."""
    msg = rfc822.Message(fp)
    id = msg.getheader('oops-id')
    exc_type = msg.getheader('exception-type')
    exc_value = msg.getheader('exception-value')
    datestr = msg.getheader('date')
    if datestr is not None:
        date =iso8601.parse_date(msg.getheader('date'))
    else:
        date = None
    pageid = msg.getheader('page-id')
    username = msg.getheader('user')
    url = msg.getheader('url')
    duration = int(float(msg.getheader('duration', '-1')))
    informational = msg.getheader('informational')
    branch_nick = msg.getheader('branch')
    revno = msg.getheader('revision')

    # Explicitly use an iterator so we can process the file
    # sequentially. In most instances the iterator will actually
    # be the file object passed in because file objects should
    # support iteration.
    lines = iter(msg.fp)

    # Request variables until the first blank line.
    req_vars = []
    for line in lines:
        line = line.strip()
        if line == '':
            break
        key, value = line.split('=', 1)
        req_vars.append((urllib.unquote(key), urllib.unquote(value)))

    # Statements until the next blank line.
    statements = []
    for line in lines:
        line = line.strip()
        if line == '':
            break
        match = re.match(
            r'^(\d+)-(\d+)(?:@([\w-]+))?\s+(.*)', line)
        assert match is not None, (
            "Unable to interpret oops line: %s" % line)
        start, end, db_id, statement = match.groups()
        if db_id is not None:
            db_id = intern(db_id)  # This string is repeated lots.
        statements.append(
            (int(start), int(end), db_id, statement))

    # The rest is traceback.
    tb_text = ''.join(lines)

    return dict(id=id, type=exc_type, value=exc_value, time=date,
            pageid=pageid, tb_text=tb_text, username=username, url=url,
            duration=duration, req_vars=req_vars, db_statements=statements,
            informational=informational, branch_nick=branch_nick, revno=revno)


def _normalise_whitespace(s):
    """Normalise the whitespace in a string to spaces"""
    if s is None:
        return None # (used by the cast to %s to get 'None')
    return ' '.join(s.split())


def _safestr(obj):
    if isinstance(obj, unicode):
        return obj.replace('\\', '\\\\').encode('ASCII',
                                                'backslashreplace')
    # A call to str(obj) could raise anything at all.
    # We'll ignore these errors, and print something
    # useful instead, but also log the error.
    # We disable the pylint warning for the blank except.
    try:
        value = str(obj)
    except:
        logging.getLogger('oops.serializer_rfc822').exception(
            'Error while getting a str '
            'representation of an object')
        value = '<unprintable %s object>' % (
            str(type(obj).__name__))
    # Some str() calls return unicode objects.
    if isinstance(value, unicode):
        return _safestr(value)
    # encode non-ASCII characters
    value = value.replace('\\', '\\\\')
    value = re.sub(r'[\x80-\xff]',
                   lambda match: '\\x%02x' % ord(match.group(0)), value)
    return value


def to_chunks(report):
    """Returns a list of bytestrings making up the serialized oops."""
    chunks = []
    chunks.append('Oops-Id: %s\n' % _normalise_whitespace(report['id']))
    if 'type' in report:
        chunks.append(
            'Exception-Type: %s\n' % _normalise_whitespace(report['type']))
    if 'value' in report:
        chunks.append(
            'Exception-Value: %s\n' % _normalise_whitespace(report['value']))
    if 'time' in report:
        chunks.append('Date: %s\n' % report['time'].isoformat())
    if 'pageid' in report:
        chunks.append(
                'Page-Id: %s\n' % _normalise_whitespace(report['pageid']))
    if 'branch_nick' in report:
        chunks.append('Branch: %s\n' % _safestr(report['branch_nick']))
    if 'revno' in report:
        chunks.append('Revision: %s\n' % report['revno'])
    if 'username' in report:
        chunks.append(
                'User: %s\n' % _normalise_whitespace(report['username']))
    if 'url' in report:
        chunks.append('URL: %s\n' % _normalise_whitespace(report['url']))
    if 'duration' in report:
        chunks.append('Duration: %s\n' % report['duration'])
    if 'informational' in report:
        chunks.append('Informational: %s\n' % report['informational'])
    chunks.append('\n')
    safe_chars = ';/\\?:@&+$, ()*!'
    if 'req_vars' in report:
        for key, value in report['req_vars']:
            chunks.append('%s=%s\n' % (urllib.quote(key, safe_chars),
                                  urllib.quote(value, safe_chars)))
        chunks.append('\n')
    if 'db_statements' in report:
        for (start, end, database_id, statement) in report['db_statements']:
            chunks.append('%05d-%05d@%s %s\n' % (
                start, end, database_id, _normalise_whitespace(statement)))
        chunks.append('\n')
    if 'tb_text' in report:
        chunks.append(report['tb_text'])
    return chunks


def write(report, output):
    """Write a report to a file."""
    output.writelines(to_chunks(report))
