# Copyright (c) 2010, 2011, Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Tests for the legacy rfc822 based [de]serializer."""

__metaclass__ = type

import datetime
import StringIO
from textwrap import dedent

from pytz import utc
import testtools

from oops.serializer_rfc822 import (
    read,
    to_chunks,
    write,
    )


class TestParsing(testtools.TestCase):

    def test_read(self):
        """Test ErrorReport.read()."""
        fp = StringIO.StringIO(dedent("""\
            Oops-Id: OOPS-A0001
            Exception-Type: NotFound
            Exception-Value: error message
            Date: 2005-04-01T00:00:00+00:00
            Page-Id: IFoo:+foo-template
            User: Sample User
            URL: http://localhost:9000/foo
            Duration: 42

            HTTP_USER_AGENT=Mozilla/5.0
            HTTP_REFERER=http://localhost:9000/
            name%3Dfoo=hello%0Aworld

            00001-00005@store_a SELECT 1
            00005-00010@store_b SELECT 2

            traceback-text"""))
        report = read(fp)
        self.assertEqual(report['id'], 'OOPS-A0001')
        self.assertEqual(report['type'], 'NotFound')
        self.assertEqual(report['value'], 'error message')
        self.assertEqual(
                report['time'], datetime.datetime(2005, 4, 1, tzinfo=utc))
        self.assertEqual(report['pageid'], 'IFoo:+foo-template')
        self.assertEqual(report['tb_text'], 'traceback-text')
        self.assertEqual(report['username'], 'Sample User')
        self.assertEqual(report['url'], 'http://localhost:9000/foo')
        self.assertEqual(report['duration'], 42)
        self.assertEqual(len(report['req_vars']), 3)
        self.assertEqual(report['req_vars'][0], ('HTTP_USER_AGENT',
                                             'Mozilla/5.0'))
        self.assertEqual(report['req_vars'][1], ('HTTP_REFERER',
                                             'http://localhost:9000/'))
        self.assertEqual(report['req_vars'][2], ('name=foo', 'hello\nworld'))
        self.assertEqual(len(report['db_statements']), 2)
        self.assertEqual(
            report['db_statements'][0],
            (1, 5, 'store_a', 'SELECT 1'))
        self.assertEqual(
            report['db_statements'][1],
            (5, 10, 'store_b', 'SELECT 2'))

    def test_read_no_store_id(self):
        """Test ErrorReport.read() for old logs with no store_id."""
        fp = StringIO.StringIO(dedent("""\
            Oops-Id: OOPS-A0001
            Exception-Type: NotFound
            Exception-Value: error message
            Date: 2005-04-01T00:00:00+00:00
            Page-Id: IFoo:+foo-template
            User: Sample User
            URL: http://localhost:9000/foo
            Duration: 42

            HTTP_USER_AGENT=Mozilla/5.0
            HTTP_REFERER=http://localhost:9000/
            name%3Dfoo=hello%0Aworld

            00001-00005 SELECT 1
            00005-00010 SELECT 2

            traceback-text"""))
        report = read(fp)
        self.assertEqual(report['id'], 'OOPS-A0001')
        self.assertEqual(report['type'], 'NotFound')
        self.assertEqual(report['value'], 'error message')
        self.assertEqual(
                report['time'], datetime.datetime(2005, 4, 1, tzinfo=utc))
        self.assertEqual(report['pageid'], 'IFoo:+foo-template')
        self.assertEqual(report['tb_text'], 'traceback-text')
        self.assertEqual(report['username'], 'Sample User')
        self.assertEqual(report['url'], 'http://localhost:9000/foo')
        self.assertEqual(report['duration'], 42)
        self.assertEqual(len(report['req_vars']), 3)
        self.assertEqual(report['req_vars'][0], ('HTTP_USER_AGENT',
                                             'Mozilla/5.0'))
        self.assertEqual(report['req_vars'][1], ('HTTP_REFERER',
                                             'http://localhost:9000/'))
        self.assertEqual(report['req_vars'][2], ('name=foo', 'hello\nworld'))
        self.assertEqual(len(report['db_statements']), 2)
        self.assertEqual(report['db_statements'][0], (1, 5, None, 'SELECT 1'))
        self.assertEqual(report['db_statements'][1], (5, 10, None, 'SELECT 2'))
        self.assertFalse(report['informational'])

    def test_read_branch_nick_revno(self):
        """Test ErrorReport.read()."""
        fp = StringIO.StringIO(dedent("""\
            Oops-Id: OOPS-A0001
            Exception-Type: NotFound
            Exception-Value: error message
            Date: 2005-04-01T00:00:00+00:00
            Page-Id: IFoo:+foo-template
            User: Sample User
            URL: http://localhost:9000/foo
            Duration: 42
            Branch: mybranch
            Revision: 45

            HTTP_USER_AGENT=Mozilla/5.0
            HTTP_REFERER=http://localhost:9000/
            name%3Dfoo=hello%0Aworld

            00001-00005@store_a SELECT 1
            00005-00010@store_b SELECT 2

            traceback-text"""))
        report = read(fp)
        self.assertEqual(report['branch_nick'], 'mybranch')
        self.assertEqual(report['revno'], '45')

    def test_minimal_oops(self):
        # If we get a crazy-small oops, we can read it sensibly.  Because there
        # is existing legacy code, all keys are filled in with None, [] rather
        # than being empty.
        fp = StringIO.StringIO(dedent("""\
            Oops-Id: OOPS-A0001
            """))
        report = read(fp)
        self.assertEqual(report['id'], 'OOPS-A0001')
        self.assertEqual(report['type'], None)
        self.assertEqual(report['value'], None)
        self.assertEqual(report['time'], None)
        self.assertEqual(report['pageid'], None)
        self.assertEqual(report['tb_text'], '')
        self.assertEqual(report['username'], None)
        self.assertEqual(report['url'], None)
        self.assertEqual(report['duration'], -1)
        self.assertEqual(len(report['req_vars']), 0)
        self.assertEqual(len(report['db_statements']), 0)
        self.assertFalse(report['informational'])
        self.assertEqual(report['branch_nick'], None)
        self.assertEqual(report['revno'], None)


class TestSerializing(testtools.TestCase):

    def test_write_file(self):
        output = StringIO.StringIO()
        report = {
            'id': 'OOPS-A0001',
            'type': 'NotFound',
            'value': 'error message',
            'time': datetime.datetime(2005, 04, 01, 00, 00, 00, tzinfo=utc),
            'pageid': 'IFoo:+foo-template',
            'tb_text': 'traceback-text',
            'username': 'Sample User',
            'url': 'http://localhost:9000/foo',
            'duration': 42,
            'req_vars': [('HTTP_USER_AGENT', 'Mozilla/5.0'),
                    ('HTTP_REFERER', 'http://localhost:9000/'),
                    ('name=foo', 'hello\nworld')],
            'db_statements': [(1, 5, 'store_a', 'SELECT 1'),
                    (5, 10, 'store_b', 'SELECT\n2')],
            'informational': False,
            'branch_nick': 'mybranch',
            'revno': '45',
            }
        write(report, output)
        self.assertEqual(output.getvalue(), dedent("""\
            Oops-Id: OOPS-A0001
            Exception-Type: NotFound
            Exception-Value: error message
            Date: 2005-04-01T00:00:00+00:00
            Page-Id: IFoo:+foo-template
            Branch: mybranch
            Revision: 45
            User: Sample User
            URL: http://localhost:9000/foo
            Duration: 42
            Informational: False

            HTTP_USER_AGENT=Mozilla/5.0
            HTTP_REFERER=http://localhost:9000/
            name%3Dfoo=hello%0Aworld

            00001-00005@store_a SELECT 1
            00005-00010@store_b SELECT 2

            traceback-text"""))

    def test_to_chunks(self):
        report = {
            'id': 'OOPS-A0001',
            'type': 'NotFound',
            'value': 'error message',
            'time': datetime.datetime(2005, 04, 01, 00, 00, 00, tzinfo=utc),
            'pageid': 'IFoo:+foo-template',
            'tb_text': 'traceback-text',
            'username': 'Sample User',
            'url': 'http://localhost:9000/foo',
            'duration': 42,
            'req_vars': [('HTTP_USER_AGENT', 'Mozilla/5.0'),
                    ('HTTP_REFERER', 'http://localhost:9000/'),
                    ('name=foo', 'hello\nworld')],
            'db_statements': [(1, 5, 'store_a', 'SELECT 1'),
                    (5, 10, 'store_b', 'SELECT\n2')],
            'informational': False,
            'branch_nick': 'mybranch',
            'revno': '45',
            }
        self.assertEqual([
            "Oops-Id: OOPS-A0001\n",
            "Exception-Type: NotFound\n",
            "Exception-Value: error message\n",
            "Date: 2005-04-01T00:00:00+00:00\n",
            "Page-Id: IFoo:+foo-template\n",
            "Branch: mybranch\n",
            "Revision: 45\n",
            "User: Sample User\n",
            "URL: http://localhost:9000/foo\n",
            "Duration: 42\n",
            "Informational: False\n",
            "\n",
            "HTTP_USER_AGENT=Mozilla/5.0\n",
            "HTTP_REFERER=http://localhost:9000/\n",
            "name%3Dfoo=hello%0Aworld\n",
            "\n",
            "00001-00005@store_a SELECT 1\n",
            "00005-00010@store_b SELECT 2\n",
            "\n",
            "traceback-text",
            ],
            to_chunks(report))

    def test_minimal_oops(self):
        # An oops with just an id, though arguably crazy, is written
        # sensibly.
        report = {'id': 'OOPS-1234'}
        self.assertEqual([
            "Oops-Id: OOPS-1234\n",
            "\n"
            ], to_chunks(report))
